/// Things used in J2D and J3D that should not be exposed to the user of the library.
#pragma once
#include <J3ML/LinearAlgebra/Vector2.hpp>
#include <JGL/types/RenderTarget.h>
#include <JGL/logger/logger.h>
#include <glad/glad.h>

namespace JGL {
    inline constexpr std::array<const LightBase*, 8> empty_light_array = { nullptr, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr };
    inline bool inJ2D = false;
    inline bool inJ3D = false;
}

namespace JGL::OpenGLState {
    inline JGL::RenderTarget* render_target = nullptr;
    inline Vector2 window_size;

    inline GLfloat oldColor[4] = {0, 0, 0, 1};
    inline GLfloat baseColor[4] = {1, 1, 1, 1};

    inline GLuint current_fbo = 0;
    inline GLint viewport[4] = {0, 0, 0, 0};

    inline bool wasTexture2DEnabled = false;
    inline bool wasTextureCoordArrayEnabled = false;
    inline bool wasNormalArrayEnabled = false;

    inline bool wasDepthTestEnabled = false;
    inline bool wasVertexArraysEnabled = false;
    inline bool wasCullFaceEnabled = false;
    inline bool wasBlendEnabled = false;
    inline bool wasColorArrayEnabled = false;
    inline GLint activeTextureUnit = 0;
}

namespace JGL::J2D {
    inline std::array<const LightBase*, 8> required_lights;
    inline std::vector<const LightBase*> light_array;
}

namespace JGL::J3D {
    // List of lights required for each objects in the scene. up-to 8. For example, the sun. Or a flash-light.
    inline std::array<const LightBase*, 8> required_lights;
    // List of all lights in the scene.
    inline std::vector<const LightBase*> optional_lights;
    inline float far_plane = 0;
    inline float fov = 0;


    // TODO handle the case that a required light is in the list of optional lights.
    inline void SelectLights(const Vector3& position) {
        std::array<const LightBase*, 8> result = {};

        unsigned int required_light_count = 0;
        for (const auto* light : required_lights)
            if (light)
                result[required_light_count++] = light;
            else
                break;

        // If there is optional lights.
        if (!optional_lights.empty()) {
            // The number of lights we need to solve for
            unsigned int remaining_lights = 8 - required_light_count;

            std::vector<std::pair<float, const LightBase*>> light_influence;
            for (const auto* light: optional_lights) {
                float influence = 0.0f;
                if (auto *point_light = dynamic_cast<const PointLight*>(light))
                    influence = point_light->GetAttenuationAtPosition(position);

                else if (auto *spot_light = dynamic_cast<const SpotLight*>(light))
                    influence = spot_light->GetAttenuationAtPosition(position);
                else
                    Logger::Warning("Couldn't determine the light type.");

                light_influence.emplace_back(influence, light);
            }

            // Sort by biggest influence.
            std::sort(light_influence.begin(), light_influence.end(),
                      [](const auto &a, const auto &b) { return a.first > b.first; });

            // Add in optional lights.
            for (unsigned int i = 0; i < remaining_lights && i < light_influence.size(); i++)
                result[required_light_count++] = light_influence[i].second;
        }

        for (unsigned int i = 0; i < result.size(); i++) {
            // If we've reached the end of an array with less than 8 lights.
            if (i > required_light_count)
                break;

            Vector4 ambient = { result[i]->GetAmbient().RN(), result[i]->GetAmbient().BN(), result[i]->GetAmbient().GN(), result[i]->GetAmbient().AN() };
            Vector4 diffuse = { result[i]->GetDiffuse().RN(), result[i]->GetDiffuse().BN(), result[i]->GetDiffuse().GN(), result[i]->GetDiffuse().AN() };
            Vector4 specular = { result[i]->GetSpecular().RN(), result[i]->GetSpecular().BN(), result[i]->GetSpecular().GN(), result[i]->GetSpecular().AN() };
            GLenum current_light = GL_LIGHT0 + i;

            glEnable(GL_LIGHT0 + i);
            glLightfv(current_light, GL_POSITION, result[i]->GetPosition().ptr());
            glLightfv(current_light, GL_AMBIENT, ambient.ptr());
            glLightfv(current_light, GL_DIFFUSE, diffuse.ptr());
            glLightfv(current_light, GL_SPECULAR, specular.ptr());
        }
    }


}